indexing

	description: "Base class that covers certain file descriptor routines."

	idea: "Functions here are also available under Windows more or less.%
	%This also means that an abstract file descriptor cannot be converted to a%
	%stream, because Windows doesn't support that."

	author: "Berend de Boer"
	date: "$Date: 2003/11/29 $"
	revision: "$Revision: #10 $"


deferred class

	ABSTRACT_FILE_DESCRIPTOR


inherit

	STDC_HANDLE [INTEGER]
		rename
			handle as fd
		redefine
			detach,
			do_close,
			is_open
		end

	EPX_CHARACTER_IO_STREAM
		rename
			end_of_input as eof,
			name as path,
			read_to_buffer as non_blocking_read_to_buffer,
			read_character as non_blocking_read_character,
			read_string as non_blocking_read_string,
			read_to_string as non_blocking_read_to_string
		redefine
			non_blocking_read_to_buffer,
			non_blocking_read_to_string,
			rewind,
			valid_unread_character
		end

	EPX_TEXT_INPUT_STREAM
		rename
			close as close_for_reading,
			end_of_input as eof,
			is_closable as is_closable_for_reading,
			name as path,
			read_to_buffer as non_blocking_read_to_buffer,
			read_character as non_blocking_read_character,
			read_string as non_blocking_read_string,
			read_to_string as non_blocking_read_to_string
		undefine
			is_closable_for_reading,
			non_blocking_read_to_buffer,
			non_blocking_read_to_string,
			rewind,
			valid_unread_character
		end

	EPX_POINTER_HELPER
		export
			{NONE} all
		end

	PORTABLE_PATH
		rename
			portable_path as path
		export
			{ANY} path
		end


feature -- Initialization

	open (a_path: STRING; a_flags: INTEGER) is
			-- Open given file with access given by `flags'.
		require
			closed: not is_open
			a_path_not_empty: a_path /= Void and then not a_path.is_empty
		do
			do_make
			do_open (a_path, a_flags)
			if is_open then
				is_open_read := True
				is_open_write:= True
			end
		ensure
			is_opened: security.error_handling.exceptions_enabled implies is_open
			open_files_increased_by_one:
				security.files.is_open_files_increased (is_owner, old security.files.open_files)
		end

	open_read (a_path: STRING) is
			-- Open given file with access given by `flags'.
		require
			closed: not is_open
			a_path_not_empty: a_path /= Void and then not a_path.is_empty
		do
			open (a_path, abstract_O_RDONLY)
			if is_open then
				is_open_write := False
			end
		ensure
			is_opened: security.error_handling.exceptions_enabled implies is_open
			open_files_increased_by_one:
				security.files.is_open_files_increased (is_owner, old security.files.open_files)
		end

	open_write (a_path: STRING) is
		require
			closed: not is_open
			a_path_not_empty: a_path /= Void and then not a_path.is_empty
		do
			open (a_path, abstract_O_WRONLY)
			if is_open then
				is_open_read := False
			end
		ensure
			is_opened: security.error_handling.exceptions_enabled implies is_open
			open_files_increased_by_one:
				security.files.is_open_files_increased (is_owner, old security.files.open_files)
		end

	open_read_write (a_path: STRING) is
		require
			closed: not is_open
			a_path_not_empty: a_path /= Void and then not a_path.is_empty
		do
			open (a_path, abstract_O_RDWR)
		ensure
			is_opened: security.error_handling.exceptions_enabled implies is_open
			open_files_increased_by_one:
				security.files.is_open_files_increased (is_owner, old security.files.open_files)
		end

	open_truncate (a_path: STRING) is
			-- Open file, if it exists, truncate it first.
		require
			closed: not is_open
			a_path_not_empty: a_path /= Void and then not a_path.is_empty
		do
			open (a_path, abstract_O_TRUNC)
			if is_open then
				is_open_read := False
			end
		ensure
			is_opened: security.error_handling.exceptions_enabled implies is_open
			open_files_increased_by_one:
				security.files.is_open_files_increased (is_owner, old security.files.open_files)
		end

	create_read_write (a_path: STRING) is
			-- Always create a file, existing or not.
			-- Give read/write permissions to user only.
		require
			closed: not is_open
			a_path_not_empty: a_path /= Void and then not a_path.is_empty
		do
			create_with_mode (
				a_path,
				abstract_O_CREAT + abstract_O_TRUNC + abstract_O_RDWR,
				abstract_S_IREAD + abstract_S_IWRITE)
		ensure
			is_opened: security.error_handling.exceptions_enabled implies is_open
			only_owner_has_read_write_permissions: True
			open_files_increased_by_one:
				security.files.is_open_files_increased (is_owner, old security.files.open_files)
		end

	create_write (a_path: STRING) is
			-- Always create a file, existing or not.
			-- Give read/write permissions to user only.
		require
			closed: not is_open
			a_path_not_empty: a_path /= Void and then not a_path.is_empty
		do
			create_with_mode (
				a_path,
				abstract_O_CREAT + abstract_O_TRUNC + abstract_O_WRONLY,
				abstract_S_IREAD + abstract_S_IWRITE)
			if is_open then
				is_open_read := False
			end
		ensure
			is_opened: security.error_handling.exceptions_enabled implies is_open
			only_owner_has_read_write_permissions: True
			open_files_increased_by_one:
				security.files.is_open_files_increased (is_owner, old security.files.open_files)
		end

	create_with_mode (a_path: STRING; flags, mode: INTEGER) is
			-- Create a file according to `flags' and with `mode' access
			-- permissions. Make sure you have th O_CREAT flag in flags
			-- if you really want to create something!
		require
			closed: not is_open
			a_path_not_empty: a_path /= Void and then not a_path.is_empty
		do
			do_make
			do_create (a_path, flags, mode)
			if is_open then
				is_open_read := True
				is_open_write := True
			end
		ensure
			is_opened: security.error_handling.exceptions_enabled implies is_open
			open_files_increased_by_one:
				security.files.is_open_files_increased (is_owner, old security.files.open_files)
		end


feature {NONE} -- Initialization

	do_make is
		require
			closed: not is_open
		do
			make_path
			clear_handle
		ensure
			path_set: path /= Void
			fd_unassigned: fd = unassigned_value
		end


feature -- Special creation

	attach_to_fd (a_fd: INTEGER; a_become_owner: BOOLEAN) is
			-- Create file descriptor with value `a_fd'. File descriptor
			-- will close it when `a_become_owner'.
		require
			closed: not is_open
			valid_file_descriptor: a_fd >= 0
		do
			do_make
			capacity := 1
			attach_to_handle (a_fd, a_become_owner)
			is_open_read := True
			is_open_write := True
		ensure
			open: is_open
			readable_and_writable: is_open_read and is_open_write
			owner_set: a_become_owner = is_owner
			open_files_increased_by_one:
				security.files.is_open_files_increased (is_owner, old security.files.open_files)
		end

	make_as_duplicate (another: ABSTRACT_FILE_DESCRIPTOR) is
			-- On creation, create a duplicate from another file descriptor
			-- As normal call, closes its own descriptor first (if open) and
			-- duplicates next.
		require
			another_file_descriptor_not_void: another /= Void
			another_open: another.is_open
		local
			old_fd,
			a_fd: INTEGER
		do
			make_path
			if is_open then
				old_fd := fd
				detach
				a_fd := abstract_dup2 (another.value, old_fd)
			else
				a_fd := abstract_dup (another.value)
			end
			if a_fd = -1 then
				raise_posix_error
			else
				capacity := 1
				set_handle (a_fd, True)
				is_open_read := True
				is_open_write := True
			end
		ensure
			open: security.error_handling.exceptions_enabled implies is_open
			own_descriptor: is_open implies is_owner
			open_files_increased_if_closed:
				memory.collecting or else
				(old (not is_open) implies
					security.files.open_files = old security.files.open_files + 1)
			open_files_not_increased_if_open:
				memory.collecting or else
				((old is_open and then old is_owner) implies
					security.files.open_files = old security.files.open_files)
		end


feature {NONE} -- Open

	do_open (a_path: STRING; flags: INTEGER) is
			-- Low level open of an existing file. `flags' should not
			-- contain the O_CREAT flag.
		require
			closed: not is_open
			a_path_not_empty: a_path /= Void and then not a_path.is_empty
			-- no_create_flag: flags && O_CREAT = 0
		local
			cpath: POINTER
			a_fd: INTEGER
		do
			set_portable_path (a_path)
			cpath := sh.string_to_pointer (path)
			a_fd := abstract_open (cpath, flags)
			sh.unfreeze_all
			if a_fd = -1 then
				raise_posix_error
			else
				capacity := 1
				set_handle (a_fd, True)
			end
		ensure
			is_opened: security.error_handling.exceptions_enabled implies is_open
			open_files_increased_by_one:
				security.files.is_open_files_increased (is_owner, old security.files.open_files)
		end

	do_create (a_path: STRING; flags, mode: INTEGER) is
			-- Low level create of the file in `a_path'.
		require
			closed: not is_open
			a_path_not_empty: a_path /= Void and then not a_path.is_empty
		local
			cpath: POINTER
			a_fd: INTEGER
		do
			set_portable_path (a_path)
			cpath := sh.string_to_pointer (path)
			a_fd := abstract_create (cpath, flags, mode)
			sh.unfreeze_all
			if a_fd = -1 then
				raise_posix_error
			else
				capacity := 1
				set_handle (a_fd, True)
			end
		ensure
			is_opened: security.error_handling.exceptions_enabled implies is_open
			open_files_increased_by_one:
				security.files.is_open_files_increased (is_owner, old security.files.open_files)
		end


feature -- Close

	detach is
			-- Forget the resource. Resource is not closed.
			-- You cannot read and write anymore.
		do
			precursor
			is_open_read := False
			is_open_write := False
		ensure then
			not_read: not is_open_read
			not_write: not is_open_write
		end


feature -- Change ownership of the descriptor. Can help to influence subtile garbage collector problems

	make_owner is
			-- This file descriptor will (start to) own its descriptor.
		obsolete "Use become_owner instead."
		do
			become_owner
		end


feature -- Read and write to memory block

	last_blocked: BOOLEAN
			-- Would last call to `read' or `write' block?

	read  (buf: POINTER; offset, nbytes: INTEGER) is
			-- Read data into `buf' at `offset' for `nbytes' bytes.
			-- The number of bytes actually read, is available in `last_read'.
		require
			open_read: is_open_read
			valid_buf: buf /= default_pointer
			valid_size: nbytes >= 0
			-- `nbytes' < SSIZE_MAX (POSIX value?)
		local
			temp_buffer: STDC_BUFFER
			remaining_bytes: INTEGER
		do
			if line_buffer = Void or else line_buffer_must_be_refilled then
				do_read (buf, offset, nbytes)
			else
				create temp_buffer.make_from_pointer (buf + offset, nbytes, False)
				remaining_bytes := line_buffer.count - line_buffer_index
				if remaining_bytes >= nbytes then
					last_read := nbytes
				else
					last_read := remaining_bytes
				end
				temp_buffer.copy_from (line_buffer, line_buffer_index, 0, last_read)
				line_buffer_index := line_buffer_index + last_read
			end
		ensure
			read_no_more_than_requested: last_read <= nbytes
		end

	write (buf: POINTER; offset, nbytes: INTEGER) is
			-- Write given data from `buf' at `offset', for `nbytes'
			-- bytes. Number of actually written bytes are in
			-- `last_written'. `last_written' can be unequal to `nbytes'
			-- if i/o is non-blocking or some error has occurred.
		require
			can_write: is_open
			valid_buf: buf /= default_pointer
			valid_size: nbytes >= 0
		local
			ptr: POINTER
			nwritten,
			nleft: INTEGER
			e: INTEGER
		do
			eof := False
			last_written := 0
			last_blocked := False
			errno.clear
			if nbytes > 0 then
				from
					nwritten := 0
					nleft := nbytes
				invariant
					nleft = 0 or else nbytes = nleft + last_written
-- `nleft' doesn't do, can remain the same in a loop.
-- Therefore no guarantee that loop terminates.
-- 				variant
-- 					nleft
				until
					nleft = 0
				loop
					ptr := buf + (offset + last_written)
					nwritten := abstract_write (fd, ptr, nleft)
					if nwritten = -1 then
						if errno.value = abstract_EWOULDBLOCK then -- test if blocked
							nleft := 0 -- loop variant + exit loop
							last_blocked := True
						elseif errno.value = abstract_EINTR then -- test if interrupted
							nwritten := 0 -- call write() again
							-- And here why the loop variant does not work.
							-- We cannot decrement `nleft'.
						else
							e := errno.value -- debug help
							raise_posix_error
							-- In case exceptions disabled, exit loop.
							nleft := 0
						end
					else
						nleft := nleft - nwritten
						last_written := last_written + nwritten
					end
				end
			else
				-- Do I still want to call write, even for zero bytes?
				-- Perhaps to get some status?
-- 				nwritten := abstract_write (fd, buf, 0)
-- 				if nwritten = -1 and then errno.value /= abstract_EWOULDBLOCK then
-- 					raise_posix_error
-- 				end
			end
		ensure
			wrote_no_more_than_requested: last_written <= nbytes
		end


feature {NONE} -- Raw read

	do_read (buf: POINTER; offset, nbytes: INTEGER) is
			-- Read data into `buf' at `offset' for `nbytes' bytes.
			-- The number of bytes actually read are available in `last_read'.
			-- Don't mix this routine with `read_string' or `read_character'!
		require
			open_read: is_open_read
			valid_buf: buf /= default_pointer
			valid_size: nbytes >= 0
			-- `nbytes' < SSIZE_MAX (POSIX value?)
		local
			ptr: POINTER
			nread,
			nleft: INTEGER
		do
			eof := False
			last_read := 0
			last_blocked := False
			errno.clear
			if nbytes > 0 then
				from
					nread := 0
					nleft := nbytes
				invariant
					nleft = 0 or else nbytes = nleft + last_read
-- Same as in `write', no guarantee this loop terminates.
-- 				variant
-- 					nleft
				until
					nleft = 0 or else last_blocked
				loop
					ptr := buf + (offset + last_read)
					nread := abstract_read (fd, ptr, nleft)
					if nread = -1 then -- test if error occurred
						if errno.value = abstract_EWOULDBLOCK then -- test if blocked
							nleft := 0
							-- Set blocked only if nothing read, else return
							-- with data just read.
							last_blocked := last_read = 0
						elseif errno.value = abstract_EINTR then -- test if interrupted
							nread := 0 -- call read() again
							-- And here is the non-terminating loop problem.
							-- `nleft' is not decremented, and we might be
							-- interrupted at every `abstract_read'.
						else
							raise_posix_error
							-- In case exceptions are disabled, exit loop.
							nleft := 0
						end
					else
						if nread = 0 then
							-- Set eof only when nothing read the first time.
							eof := nleft = nbytes
							-- Satisfy loop invariant, finish loop.
							nleft := 0
						else
							last_read := last_read + nread
							-- Do not read more than what is available,
							-- else application might block.
							nleft := 0
						end
					end
				end
			else
				-- Still call read() for zero bytes?
-- 				nread := abstract_read (fd, buf, 0)
-- 				if nread = -1 and then errno.value /= abstract_EWOULDBLOCK then
-- 					raise_posix_error
-- 				end
			end
		ensure
			read_no_more_than_requested: last_read <= nbytes
		end


feature -- Read/write to/from STDC_BUFFER

	read_buffer (buf: STDC_BUFFER; offset, nbytes: INTEGER) is
			-- Read data into `buf' at `offset' for `nbytes' bytes.
			-- Number of bytes actually read are available in `last_read'.
			-- Thisis a more safe version of `read' in case you have a
			-- STDC_BUFFER object.
		do
			read (buf.ptr, offset, nbytes)
		end

	put_buffer, write_buffer (buf: STDC_BUFFER; offset, nbytes: INTEGER) is
			-- More safe version of `write' in case you have a
			-- STDC_BUFFER object.
		do
			write (buf.ptr, offset, nbytes)
		end


feature -- Eiffel like output

	put (a: ANY) is
			-- Write any Eiffel object as string using its `out' value.
		do
			if a /= Void then
				put_string (a.out)
			end
		end

	put_character (c: CHARACTER) is
			-- Write a character.
		do
			temp_character := c
			write ($temp_character, 0, 1)
		end

	write_character (c: CHARACTER) is
			-- Write a character.
		require
			is_open_write: is_open_write
		do
			put_character (c)
		end

	write_string, puts (a_string: STRING) is
			-- Write `a_string' to output stream.
		require
			is_open_write: is_open_write
			a_string_not_void: a_string /= Void
			no_null_character: not a_string.has ('%U')
		do
			put_string (a_string)
		end

	put_string (a_string: STRING) is
			-- Write `a_string' to output stream.
		local
			p: POINTER
			uc: UC_STRING
		do
			uc ?= a_string
			if uc /= Void then
				p := sh.uc_string_to_pointer (uc)
				write (p, 0, uc.byte_count)
			else
				p := sh.string_to_pointer (a_string)
				write (p, 0, a_string.count)
			end
			sh.unfreeze_all
		end


feature -- Buffered input

	last_character: CHARACTER
			-- Last character read

	last_line: STRING is
		obsolete "Use read_string instead."
		do
			Result := last_string
		end

	non_blocking_read_to_buffer (a_buffer: KI_BUFFER [CHARACTER]; pos, nb: INTEGER): INTEGER is
			-- Fill `a_buffer', starting at position `pos', with
			-- at most `nb' items read from input stream.
			-- Return the number of items actually read.
		local
			blocking_disabled: BOOLEAN
		do
			blocking_disabled := not is_blocking_io
			if blocking_disabled then
				set_blocking_io (True)
			end
			Result := precursor (a_buffer, pos, nb)
			if blocking_disabled then
				blocking_disabled := False
				set_blocking_io (False)
			end
		rescue
			if blocking_disabled then
				set_blocking_io (False)
			end
		end

	non_blocking_read_character is
			-- Read the next item in input stream.
			-- Make the result available in `last_item'.
		local
			blocking_disabled: BOOLEAN
		do
			blocking_disabled := not is_blocking_io
			if blocking_disabled then
				set_blocking_io (True)
			end
			read_character
			if blocking_disabled then
				blocking_disabled := False
				set_blocking_io (False)
			end
		rescue
			if blocking_disabled then
				set_blocking_io (False)
			end
		end

	non_blocking_read_string (nb: INTEGER) is
			-- Read at most `nb' characters from input stream.
			-- Make the characters that have actually been read
			-- available in `last_string'.
		local
			blocking_disabled: BOOLEAN
		do
			blocking_disabled := not is_blocking_io
			if blocking_disabled then
				set_blocking_io (True)
			end
			read_string (nb)
			if blocking_disabled then
				blocking_disabled := False
				set_blocking_io (False)
			end
		rescue
			if blocking_disabled then
				set_blocking_io (False)
			end
		end

	non_blocking_read_to_string (a_string: STRING; pos, nb: INTEGER): INTEGER is
			-- Fill `a_string', starting at position `pos', with
			-- at most `nb' characters read from input stream.
			-- Return the number of characters actually read.
		local
			blocking_disabled: BOOLEAN
		do
			blocking_disabled := not is_blocking_io
			if blocking_disabled then
				set_blocking_io (True)
			end
			Result := precursor (a_string, pos, nb)
			if blocking_disabled then
				blocking_disabled := False
				set_blocking_io (False)
			end
		rescue
			if blocking_disabled then
				set_blocking_io (False)
			end
		end

	read_character is
			-- Sets `last_character'.
			-- If this routine blocks, `last_character' has the value
			-- '%U'. Therefore, if non-blocking is enabled, always check
			-- `last_blocked' to see if the value make sense.
		require
			is_open_read: is_open_read
			not_end_of_input: not eof
		do
			assert_have_line_buffer
			if line_buffer_must_be_refilled then
				read_buffer (line_buffer, 0, line_buffer.capacity)
				line_buffer.set_count (last_read)
				line_buffer_index := 0
			end
			if line_buffer_index < line_buffer.count then
				last_character := line_buffer.peek_character (line_buffer_index)
				line_buffer_index := line_buffer_index + 1
			else
				last_character := '%U'
			end
		ensure
			last_character_reset_on_eof: eof implies last_character = '%U'
			zero_character_on_block: last_blocked implies last_character = '%U'
		end

	read_line is
			-- Read characters from input stream until a line separator
			-- or end of file is reached. Make the characters that have
			-- been read available in `last_string' and discard the line
			-- separator characters from the input stream.
			-- Zero characters will be read when non-blocking i/o
			-- is enabled, and `read' would block.
		do
			-- Make sure last_string has at least `nb' capacity.
			if last_string = Void then
				create last_string.make (1024)
			else
				sh.wipe_out (last_string)
			end

			-- Read up to newline or eof, whatever comes first.
			-- Routine probably can be improved somewhat by reading from
			-- line buffer directly and copying data in blocks instead of
			-- character by character.
			from
				read_character
			until
				last_character = '%N' or else
				eof or else last_blocked
			loop
				last_string.append_character (last_character)
				read_character
			end

			-- Have to remove %R character?
			if last_character = '%N' then
				-- We've read at least one character, so don't return `eof'.
				-- Note that `last_string' might be empty, because we
				-- remove the end-of-line character.
				eof := False
				if last_string.count > 0 and last_string.item (last_string.count) = '%R' then
					last_string.remove (last_string.count)
				end
			else
				-- Line is empty or line does not end with end-of-line
				-- character. If read something, don't return eof.
				if eof and then not last_string.is_empty then
					eof := False
				end
			end
		end

	read_new_line is
			-- Read a line separator from input file.
			-- Make the characters making up the recognized
			-- line separator available in `last_string',
			-- or make `last_string' empty and leave the
			-- input file unchanged if no line separator
			-- was found.
		do
			-- Not supported, so we never find a new line.
			last_string.wipe_out
		end

	read_string (nb: INTEGER) is
			-- Read at most `nb' characters from input stream.
			-- Make the characters that have actually been read
			-- available in `last_string'.
			-- Zero characters will be read when non-blocking i/o
			-- is enabled, and `read' would block.
		require
			is_open_read: is_open_read
			not_end_of_input: not eof
			nb_large_enough: nb > 0
		do
			-- Make sure last_string has at least `nb' capacity.
			if last_string = Void then
				if nb > 1024 then
					create last_string.make_filled (' ', nb)
				else
					create last_string.make_filled (' ', 1024)
				end
			else
				if last_string.count < nb then
					-- Have to use create, its only a creation routine in VE
					create last_string.make_filled (' ', nb)
				end
			end

			-- Read up to newline or `nb' or eof, whatever comes first.
			-- Routine probably can be improved somewhat by reading from
			-- line buffer directly and copying data in blocks instead of
			-- character by character.
			-- This implementation just calls read() with a pointer to
			-- the string area. I assume that when calling `read_string'
			-- in a loop, it should work quite well.

			-- But because the bytes read can vary, sometimes we have to
			-- reallocate the string. And that might be bad for
			-- performance. In that case a single fixed buffer would work
			-- better, but we would have to copy the string from that
			-- buffer to `last_string'.
			read (sh.string_to_pointer (last_string), 0, nb)
			sh.unfreeze_all
			last_string.keep_head (last_read)

			if eof and then not last_string.is_empty then
				eof := False
			end
		ensure
			last_string_not_void: not eof implies last_string /= Void
			last_string_count_small_enough: last_string.count <= nb
			eof_only_when_nothing_read:
				not last_blocked implies last_string.is_empty = eof
			empty_only_when_blocked:
				not eof implies last_blocked = last_string.is_empty
		end


	-- This is the naive implementation that will *not* work for sockets
	-- because it tries to read up to `nb' characters before returning.
	-- The best way is to read a block and return immediately after
	-- that block has been read.

-- 	read_string (nb: INTEGER) is
-- 			-- Read at most `nb' characters from input stream.
-- 			-- Make the characters that have actually been read
-- 			-- available in `last_string'.
-- 			-- Zero characters will be read when non-blocking i/o
-- 			-- is enabled, and `read' would block.
-- 		require
-- 			is_open_read: is_open_read
-- 			not_end_of_input: not eof
-- 			nb_large_enough: nb > 0
-- 		do
-- 			-- Make sure last_string has at least `nb' capacity.
-- 			if last_string = Void then
-- 				if nb > 1024 then
-- 					create last_string.make ( nb)
-- 				else
-- 					create last_string.make (1024)
-- 				end
-- 			else
-- 				last_string.wipe_out
-- 			end

-- 			-- Read up to newline or `nb' or eof, whatever comes first.
-- 			-- Routine probably can be improved somewhat by reading from
-- 			-- line buffer directly and copying data in blocks instead of
-- 			-- character by character.
-- 			from
-- 			until
-- 				last_string.count = nb or else eof or else last_blocked
-- 			loop
-- 				read_character
-- 				if not eof and then not last_blocked then
-- 					last_string.append_character (last_character)
-- 				end
-- 			end

-- 			if eof and then not last_string.is_empty then
-- 				eof := False
-- 			end
-- 		ensure
-- 			last_string_not_void: not eof implies last_string /= Void
-- 			last_string_count_small_enough: last_string.count <= nb
-- 			eof_only_when_nothing_read:
-- 				not last_blocked implies last_string.is_empty = eof
-- 			empty_only_when_blocked:
-- 				not eof implies last_blocked = last_string.is_empty
-- 		end


feature {NONE} -- Buffered reading support

	line_buffer: EPX_SMART_BUFFER
			-- Buffer used in `read_character' and `read_string'.

	line_buffer_index: INTEGER
			-- Offset to unread content in `line_buffer'.

	line_buffer_must_be_refilled: BOOLEAN is
			-- Does `line_buffer_index' indicate that the end of
			-- `line_buffer' is reached?
		require
			have_line_buffer: line_buffer /= Void
		do
			Result := line_buffer_index >= line_buffer.count
		end

	assert_have_line_buffer is
		do
			if line_buffer = Void then
				create line_buffer.allocate (1024)
			end
		ensure
			line_buffer_not_void: line_buffer /= Void
		end


feature -- Unreading

	valid_unread_character (a_character: CHARACTER): BOOLEAN is
			-- Can `a_character' be put back in input stream?
		do
			-- Unreading not supported.
			Result := False
		end


feature {NONE} -- Unreading

	unread_character (an_item: CHARACTER) is
		do
			-- ignore, not supported.
			-- We can support it, and should, else the implementation of
			-- `read_new_line' is incorrect. Just stuff things in `line_buffer'.
		end


feature -- File position

	rewind is
			-- Move input positionto the beginning of stream.
		do
			seek (0)
		end

	seek (offset: INTEGER) is
			-- Set file position to given absolute `offset'.
		require
			valid_offset: offset >= 0
		do
			safe_call (abstract_lseek (fd, offset, SEEK_SET))
		end

	seek_from_current (offset: INTEGER) is
			-- Set file position relative to current position.
		do
			safe_call (abstract_lseek (fd, offset, SEEK_CUR))
		end

	seek_from_end (offset: INTEGER) is
			-- Set file position relative to end of file.
		require
			valid_offset: offset <= 0
		do
			safe_call (abstract_lseek (fd, offset, SEEK_END))
		end


feature -- Status report

	eof: BOOLEAN
			-- True if end-of-file reached.
			-- Currently I'm unsure if detection is reliable.

	isatty: BOOLEAN is
		obsolete "Use is_attached_to_terminal instead."
		do
			Result := is_attached_to_terminal
		end

	is_attached_to_terminal: BOOLEAN is
			-- Is the handle associated with character device?
		require
			open: is_open
		do
			Result := abstract_isatty (fd)
		end

	is_closed: BOOLEAN is
			-- Is file descriptor is closed?
		obsolete "Use not is_open instead."
		do
			Result := not is_open
		ensure
			in_balance: Result implies not is_open
		end

	is_open: BOOLEAN is
			-- Does `handle' contain an open handle?
		do
			-- Also return False of path = Void. When object is
			-- just created, handle = 0, so is_open would return True.
			Result :=
				fd /= unassigned_value and then
				path /= Void
		end

	resource_usage_can_be_increased: BOOLEAN is
			-- Is it allowed to open another file?
		do
			Result := security.files.open_files <= security.files.max_open_files
		end


feature -- Access

	status: ABSTRACT_STATUS is
			-- The status for this file descriptor. Cached value,
			-- refreshed only when file reopened.
		require
			open: is_open
		do
			if my_status = Void then
				make_status
			end
			Result := my_status
		ensure
			status_not_void: status /= Void
		end

	value: INTEGER is
			-- The actual file descriptor value.
		require
			valid_file_descriptor: is_open
		do
			Result := fd
		end


feature -- non-blocking i/o

	is_blocking_io: BOOLEAN is
			-- Is blocking i/o enabled?
			-- Blocking i/o is the default.
			-- If false, calls like `read' and `write' will never wait
			-- for input, if there is no input.
		require
			open: is_open_read
		do
			Result := True
		end

	set_blocking_io (enable: BOOLEAN) is
			-- Set `is_blocking_io'.
		require
			supports_nonblocking_io: not enable implies supports_nonblocking_io
			open: is_open
		do
			-- do nothing
		ensure
			blocking_set: enable = is_blocking_io
		end

	supports_nonblocking_io: BOOLEAN is
			-- Does this descriptor support non-blocking input/output?
			-- On POSIX systems, any descriptor does.
			-- On Windows sockets and pipes do.
		do
			Result := False
		end


feature -- Gobo stream specifics

	flush is
		do
			-- no meaning
		end


feature {NONE} -- Implementation

	my_status: ABSTRACT_STATUS
			-- Cached status object.

	clear_fd is
			-- Only allowed routine that may reset a file descriptor.
		obsolete "Use clear_handle instead."
		do
			clear_handle
		end

	make_status is
			-- Give `my_status' a proper value.
		require
			open: is_open
		deferred
		ensure
			status_is_set: my_status /= Void
		end

	temp_character: CHARACTER
			-- Needed by SmartEiffel to be able to pass a pointer to it.


feature {NONE} -- Low level handle functions

	do_close: BOOLEAN is
			-- Close resource. Return False if an error occurred. Error
			-- value should be in `errno'. This routine may never call
			-- another object, else it cannot be used safely in
			-- `dispose'.
			-- This routine is usely redefined to actually close or
			-- deallocate the resource in addition of resetting `handle'.
		local
			r: INTEGER
		do
			r := abstract_close (fd)
			Result := r /= -1
			clear_handle
			my_status := Void
			is_open_read := False
			is_open_write := False
		end

	unassigned_value: INTEGER is
			-- The value that indicates that `handle' is unassigned.
		do
			Result := -1
		end


feature {NONE} -- Counting of allocated resource

	decrease_open_files is
			-- Decreate the number of open files, does not access any
			-- other object, so can be called from `dispose' if one
			-- inherits from this class.
		external "C"
		alias "posix_decrease_open_files"
		end

	decrease_resource_count is
			-- Decrease number of allocated resource by `capacity'.
			-- Due to limitations of certain Eiffel implementations, this
			-- routine may not call any other object. Calling C code is safe.
		do
			decrease_open_files
		end

	increase_resource_count is
			-- Increase number of allocated resources by `capacity'.
		do
			security.files.increase_open_files
		end

	resource_count: INTEGER is
			-- Currently allocated number of resources. It's a global
			-- number, counting the `capacity' of all owned resources of
			-- this type.
		do
			Result := security.files.open_files
		end


feature {NONE} -- Abstract API binding

	abstract_close (fildes: INTEGER): INTEGER is
			-- Closes a a file. Returns -1 on error.
		deferred
		end

	abstract_create (a_path: POINTER; oflag, mode: INTEGER): INTEGER is
			-- Creates a new file or rewrites an existing one
			-- does not call `creat' but `open'!
		deferred
		end

	abstract_dup (fildes: INTEGER): INTEGER is
			-- Duplicates an open file descriptor.
		deferred
		ensure
			-- Result = -1 implies errno.value is set
		end

	abstract_dup2 (fildes, fildes2: INTEGER): INTEGER is
			-- Duplicates an open file descriptor.
			-- fildes is the file descriptor to duplicate, fildes2 is the
			-- desired new file descriptor.
		deferred
		end

	abstract_isatty (fildes: INTEGER): BOOLEAN is
			-- Determines if a file descriptor is associated with a terminal.
		deferred
		end

	abstract_lseek (fildes: INTEGER; offset, whence: INTEGER): INTEGER is
			-- Repositions read/write file offset.
		deferred
		end

	abstract_open (a_path: POINTER; oflag: INTEGER): INTEGER is
			-- Open a file.
		require
			a_path_not_null: a_path /= default_pointer
		deferred
		end

	abstract_read (fildes: INTEGER; buf: POINTER; nbyte: INTEGER): INTEGER is
			-- Read from a file.
		require
			valid_fildes: fildes /= unassigned_value
			valid_buf: buf /= default_pointer
			valid_bytes: nbyte >= 0
		deferred
		ensure
			-- Result = -1 implies errno.value is set
		end

	abstract_write (fildes: INTEGER; buf: POINTER; nbyte: INTEGER): INTEGER is
			-- Write to a file.
		require
			valid_fildes: fildes /= unassigned_value
			valid_buf: buf /= default_pointer
			valid_bytes: nbyte >= 0
		deferred
		ensure
			-- Result = -1 implies errno.value is set
		end


feature {NONE} -- Error codes

	abstract_EWOULDBLOCK: INTEGER is
			-- The process would be delayed in the I/O operation.
		deferred
		end

	abstract_EINTR: INTEGER is
			-- Read/write was interrupted.
		deferred
		end


feature {NONE} -- Open constants

	abstract_O_APPEND: INTEGER is
			-- Set the file offset to the end-of-file prior to each write.
		deferred
		end

	abstract_O_CREAT: INTEGER is
			-- If the file does not exist, allow it to be created. This
			-- flag indicates that the mode argument is present in the
			-- call to open.
		deferred
		end

	abstract_O_RDONLY: INTEGER is
			-- Open for reading only.
		deferred
		end

	abstract_O_RDWR: INTEGER is
			-- Open for reading and writing.
		deferred
		end

	abstract_O_TRUNC: INTEGER is
			-- Use only on ordinary files opened for writing. It causes
			-- the file to be truncated to zero length.
		deferred
		end

	abstract_O_WRONLY: INTEGER is
			-- Open for writing only.
		deferred
		end


feature {NONE} -- Permission constants

	abstract_S_IREAD: INTEGER is
		deferred
		end

	abstract_S_IWRITE: INTEGER is
		deferred
		end


invariant

	valid_status: not is_open implies my_status = Void
	path_not_void: path /= Void
	line_buffer_index_offset_ok: line_buffer /= Void implies line_buffer_index <= line_buffer.count

end
